#!/usr/bin/env python3
import os
import sys
import json
import argparse
import concurrent.futures

from tqdm import tqdm
from functools import partial

from .polipy import *
from .constants import UTC_DATE, CWD
from .logger import get_logger

def _setup_parser():
    """
    Sets up the CLI argument parser.
    """
    parser = argparse.ArgumentParser(description='Download privacy policies from URLs contained in the `input_file`.')
    parser.add_argument('input_file', help='Path to file containing a list of newline-separated URLs of privacy policies to scrape.')
    parser.add_argument('--output_dir', '-o', default=CWD, help='Path to directory where policies will be saved (default is {}).'.format(CWD))
    parser.add_argument('--timeout', '-t', default=30, type=int, help='The amount of time in seconds to wait for the HTTP request response (default is 30).')
    parser.add_argument('--screenshot', '-s', action='store_true', help='Capture and save the screenshot of the privacy policy page (default is False).')
    parser.add_argument('--extractors', '-e', nargs='+', default=['text'], help='Extractors to use to capture information from the privacy policy (default is ["text"]).')
    parser.add_argument('--workers', '-w', default=1, type=int, help='Number of threading workers to use (default is 1).')
    parser.add_argument('--force', '-f', action='store_true', help='Scrape privacy policy again even if it is already scraped or has not been updated (default is False).')
    parser.add_argument('--raise_errors', '-r', action='store_true', help='Raise errors that occur during the scraping and parsing (default is False).')
    parser.add_argument('--verbose', '-v', action='store_true', help='Enable verbose logging (default is False).')
    return parser

class CommandLineTool:

    def __init__(self):
        """
        Sets up the command line tool and handles logging and argument parsing.
        """
        parser = _setup_parser()
        args = parser.parse_args()
        vargs = {k:v for k, v in vars(args).items() if v is not None}
        vargs['logger'] = get_logger(args.verbose)
        self.policies(**vargs)

    def policies(self, input_file, workers, **kwargs):
        """
        Loads the provided list of URLs from the `input_file`, obtains the
        requested information for each policy and saves the
        resulting file to a directory in `output_dir` (defaults to current working directory).
        """
        # Load the list of policy URLs.
        with open(input_file, 'r') as f:
            urls = set([x.strip() for x in f.read().strip().split()])

        with concurrent.futures.ThreadPoolExecutor(max_workers=workers) as executor:
            futures = {executor.submit(download_policy, url, **kwargs) for url in urls}
            list(tqdm(concurrent.futures.as_completed(futures), total=len(futures)))


def main():
    CommandLineTool()

if __name__ == "__main__":
    main()



from bs4 import BeautifulSoup
from io import BytesIO, StringIO
from pdfminer.high_level import extract_text as parse_pdf

extractors = [
    'text'
]

def extract(extractor, **kwargs):
    if extractor == 'text':
        content = extract_text(**kwargs)
    return content

def extract_text(url_type, url=None, dynamic_source=None, static_source=None, **kwargs):
    if url_type is None or url_type in ['html', 'other']:
        content = extract_html(dynamic_source, url)
    elif url_type == 'pdf':
        content = extract_pdf(static_source)
    elif url_type == 'plain':
        content = dynamic_source
    else:
        print(url_type, url)
        content = dynamic_source
    return content

def extract_pdf(source):
    f = BytesIO(source.encode())
    text = parse_pdf(f)
    return text.strip()

def extract_html(source, url):
    if url == 'docs.google.com' and '"s":"' in source:
        return extract_google_docs(source)
    else:
        return extract_other(source)

def extract_google_docs(source):
    chunks = source.split('"s":"')[1:]
    chunks = [x.split('"},')[0].replace('\\n', '\n').replace('\\u000b', '\n') for x in chunks]
    text = ''.join(chunks).strip()
    return text

def extract_other(source):
    """
    From https://stackoverflow.com/questions/328356/extracting-text-from-html-file-using-python
    """
    soup = BeautifulSoup(source, 'html.parser')

    for script in soup(['script','style']):
        script.extract()

    text = soup.get_text(separator=' ')
    lines = (line.strip() for line in text.splitlines())
    chunks = (phrase.strip() for line in lines for phrase in line.split('  '))
    text = '\n'.join(chunk for chunk in chunks if chunk)
    return text



from .networking import get_url_type, parse_url, scrape
from .extractors import extract, extractors
from .constants import UTC_DATE, CWD
from .exceptions import NetworkIOException, ParserException
from .logger import get_logger

import os
import json
import pathlib
import hashlib
import logging

# Public module class.
class Policy:
    """
    A class representing a privacy policy.

    Attributes
    ----------
    url : dict
        Contains the URL to the privacy policy and additional information about the URL, such as domain, scheme, content-type, etc.
    source : dict
        Contains the information scraped from the webpage where policy is hosted, such as the HTML (dynamic) source and the screenshot.
    content : dict
        Contains the content extracted from the privacy policy website, such as the text of the policy.
    """

    url, source, content = {}, {}, {}

    def __init__(self, url):
        """
        Constructor method. Populates the `Policy.url` attribute.

        Parameters
        ----------
        url : str
            The URL of the privacy policy.
        """
        self.url['url'] = url
        self.url = self.url | parse_url(url)
        self.url['domain'] = self.url['domain'].strip().strip('.').strip('/')

        # Generate the hash to avoid collisions in output file names.
        self.url['hash'] = hashlib.md5(url.encode()).hexdigest()[:10]

        # Define the output directory name for this policy.
        d, h = self.url['domain'].replace('.', '_'), self.url['hash']
        self.output_dir = '{}_{}'.format(d, h)

    def scrape(self, timeout=30, **kwargs):
        """
        Obtains the page source of the given privacy policy URL.

        Populates the `Policy.source` attribute.

        Parameters
        ----------
        screenshot : bool, optional
            Flag that indicates whether to capture and save the screenshot of the privacy policy page (default is `False`).
        timeout : int, optional
            The amount of time in seconds to wait for the HTTP request response (default is `30`).
        **kwargs : dict
            Additional keyword arguments.

        Returns
        -------
        polipy.Policy
            `Policy` object with the populated attribute.

        Raises
        ------
        NetworkIOException
            Raised if an error occured while performing networking I/O.
        """
        self.url['type'], self.source['static_html'] = get_url_type(self.url['url'], timeout)
        if self.url['type'] in ['html', 'pdf']:
            self.source = self.source | scrape(self.url['url'], timeout=timeout, **kwargs)
        else:
            self.source['dynamic_html'] = self.source['static_html']
        return self

    def extract(self, extractors=['text'], **kwargs):
        """
        Extracts information from the scraped privacy policy.

        Populates the `Policy.content` attribute.

        Parameters
        ----------
        extractors : list of str, optional
            Extractors to use to capture information from the privacy policy (default is ["text"]).
        **kwargs : dict
            Additional keyword arguments.

        Returns
        -------
        polipy.Policy
            `Policy` object with the populated attribute.

        Raises
        ------
        ParserException
            Raised if an error occured while extracting text from page source.
        """
        vargs = {
            'url': self.url['url'],
            'url_type': self.url['type'],
            'static_source': self.source['static_html'],
            'dynamic_source': self.source['dynamic_html'],
            **kwargs
        }
        for extractor in extractors:
            content = extract(extractor, **vargs)
            self.content[extractor] = content
        return self

    def save(self, output_dir, **kwargs):
        """
        Saves the information contained in the `Policy` object using the following directory structure:

        ├── <output_dir>
        │   ├── <policy URL domain>_<hash of policy URL>
        │   │   ├── <current UTC date>.html
        │   │   ├── <current UTC date>.json
        └── └── └── <current UTC date>.meta

        Parameters
        ----------
        output_dir : str
            Path to directory where the policy will be saved.
        **kwargs : dict
            Additional keyword arguments.
        """
        # Create the output directory if it does not exist.
        policy_output_dir = os.path.join(output_dir, self.output_dir)
        pathlib.Path(policy_output_dir).mkdir(parents=True, exist_ok=True)

        # Define output formats.
        output = {
            'html': os.path.join(policy_output_dir, '{}.{}'.format(UTC_DATE, 'html')),
            'png': os.path.join(policy_output_dir, '{}.{}'.format(UTC_DATE, 'png')),
            'json': os.path.join(policy_output_dir, '{}.{}'.format(UTC_DATE, 'json')),
            'pdf': os.path.join(policy_output_dir, '{}.{}'.format(UTC_DATE, 'pdf')),
            'txt': os.path.join(policy_output_dir, '{}.{}'.format(UTC_DATE, 'txt')),
            'meta': os.path.join(policy_output_dir, '{}.{}'.format(UTC_DATE, 'meta')),
        }

        meta = {'last_scraped': UTC_DATE} | self.url

        # Save results.
        if 'dynamic_html' in self.source and self.source['dynamic_html'] is not None:
            with open(output['html'], 'w') as f:
                f.write(self.source['dynamic_html'])
        if 'png' in self.source and self.source['png'] is not None:
            with open(output['png'], 'wb') as f:
                f.write(self.source['png'])
        if len(self.content) > 0:
            with open(output['json'], 'w') as f:
                json.dump(self.content, f)
        if self.url['type'] == 'pdf' and 'static_html' in self.source and self.source['static_html'] is not None:
            with open(output['pdf'], 'w') as f:
                f.write(self.source['static_html'])
        if self.url['type'] == 'plain' and 'static_html' in self.source and self.source['static_html'] is not None:
            with open(output['txt'], 'w') as f:
                f.write(self.source['static_html'])
        if len(self.url) > 0:
            with open(output['meta'], 'w') as f:
                json.dump(meta, f)

    def to_dict(self):
        """
        Converts the `Policy` object to a dictionary.

        Returns
        -------
        dict
            Dictionary containing policy attributes as key-value pairs.
        """
        obj = {
            'url': self.url,
            'source': self.source,
            'content': self.content
        }
        return obj

    def __repr__(self):
        return '{}({})'.format(self.__class__, self.to_dict())

# Public module methods.
def get_policy(url, screenshot=False, timeout=30, extractors=['text'], **kwargs):
    """
    Helper method that returns a `polipy.Policy` object containing
    information about the policy, scraped and processed from the given URL.

    Parameters
    ----------
    url : str
        The URL of the privacy policy.
    screenshot : bool, optional
        Flag that indicates whether to capture and save the screenshot of the privacy policy page (default is `False`).
    timeout : int, optional
        The amount of time in seconds to wait for the HTTP request response (default is `30`).
    extractors : list of str, optional
        Extractors to use to capture information from the privacy policy (default is `["text"]`).
    **kwargs : dict
        Additional keyword arguments.

    Returns
    -------
    polipy.Policy
        Object containing information about the privacy policy.
        Please view documentation for `polipy.Policy` class for more information.

    Raises
    ------
    NetworkIOException
        Raised if an error occured while performing networking I/O.

    ParserException
        Raised if an error occured while extracting text from page source.
    """
    policy = Policy(url)
    policy.scrape(screenshot=screenshot, timeout=timeout)
    policy.extract(extractors=extractors)
    return policy

def download_policy(url, output_dir=CWD, force=False, raise_errors=False, logger=None, **kwargs):
    """
    Helper method that scrapes, parses, and saves the privacy policy located at the provided `url`
    by creating the following directory structure:

    ├── <output_dir>
    │   ├── <policy URL domain>_<hash of policy URL>
    │   │   ├── <current UTC date>.html
    │   │   ├── <current UTC date>.json
    └── └── └── <current UTC date>.meta

    Parameters
    ----------
    url : str
        The URL of the privacy policy.
    output_dir : str, optional
        Path to directory where the policy will be saved (default is the current working directory).
    force : bool, optional
        Flag that indicates whether to scrape privacy policy again even if it is already scraped or has not been updated (default is `False`).
    raise_errors : bool, optional
        Flag that indicates whether to raise errors that occur during the scraping and parsing (default is `False`).
    logger : logging.Logger, optional
        A `logging.Logger` object to handle the logging of events (default is `None`).
    screenshot : bool, optional
        Flag that indicates whether to capture and save the screenshot of the privacy policy page (default is `False`).
    timeout : int, optional
        The amount of time in seconds to wait for the HTTP request response (default is `30`).
    extractors : list of str, optional
        Extractors to use to capture information from the privacy policy (default is `["text"]`).
    **kwargs : dict
        Additional keyword arguments.

    Raises
    ------
    NetworkIOException
        Raised if an error occured while performing networking I/O.

    ParserException
        Raised if an error occured while extracting text from page source.
    """
    # Get handle on the logger object.
    logger = logger if logger is not None else logging.getLogger('polipy')

    # Get policy object with the associated URL.
    policy = Policy(url)

    # Get a handle on the output directory name for this policy.
    policy_output_dir = os.path.join(output_dir, policy.output_dir)
    files = os.listdir(policy_output_dir) if os.path.isdir(policy_output_dir) else []

    # Skip scraping if policy was already scraped today unless force flag is set.
    N = len([x for x in files if x.split('.')[0] == UTC_DATE])
    if N > 0 and not force:
        logger.warning('Privacy policy was already scraped today from {} -- skipping.'.format(url))
        return

    # Scrape the policy.
    try:
        policy.scrape(**kwargs)
    except NetworkIOException as e:
        if raise_errors:
            raise e from None
        logger.warning('Encountered a network exception while scraping policy from {} -- skipping.\n{}'.format(url, e))
        return

    # Extract information from policy.
    policy.extract(**kwargs)

    # Skip saving the file if the policy did not change unless force flag is set.
    N = len([x for x in files if x.endswith('.json')])
    if N > 0 and not force:
        latest_file = sorted(files, key=lambda x: int(x.split('.')[0]))[-1]
        with open(os.path.join(policy_output_dir, latest_file), 'r') as f:
            last_policy = json.load(f)
        if 'text' in last_policy and (last_policy['text'] == policy.content['text']):
            logger.warning('Privacy policy has not changed since {} for {} -- skipping.'.format(latest_file.split('.')[0], url))
            return

    logger.info('Saving privacy policy obtained from {} to {}.'.format(url, os.path.abspath(policy_output_dir)))
    policy.save(output_dir=output_dir)



import requests
from urllib.parse import urlparse
from .exceptions import NetworkIOException

from selenium import webdriver
from selenium.webdriver.firefox.options import Options
from selenium.common.exceptions import TimeoutException, WebDriverException

def get_url_type(url, timeout):
    """
    https://stackoverflow.com/questions/38690586/determine-if-url-is-a-pdf-or-html-file
    """
    try:
        r = requests.get(url, timeout=timeout)
    except requests.exceptions.RequestException as e:
        raise NetworkIOException(e) from None
    content_type = r.headers.get('content-type', '')
    url_type = None
    if 'application/pdf' in content_type:
        url_type = 'pdf'
    elif 'text/html' in content_type:
        url_type = 'html'
    elif 'text/plain' in content_type:
        url_type = 'plain'
    else:
        url_type = 'other'
    return url_type, r.content.decode(errors='ignore')

def parse_url(url):
    parsed = urlparse(url)
    result = {
        'scheme': parsed.scheme,
        'domain': parsed.netloc,
        'path': parsed.path,
        'params': parsed.params,
        'query': parsed.query,
        'fragment': parsed.fragment,
    }
    return result

def scrape(url, screenshot, timeout, **kwargs):
    options = Options()
    options.add_argument('-headless')
    options.add_argument('-private')
    driver = webdriver.Firefox(options=options)
    driver.set_page_load_timeout(timeout)
    response = {}
    try:
        driver.get(url)
        response['dynamic_html'] = driver.page_source
        response['png'] = driver.get_screenshot_as_png() if screenshot else None
    except (TimeoutException, WebDriverException) as e:
        raise NetworkIOException(e) from None
    finally:
        driver.quit()
    return response



import logging, sys

def get_logger(verbose=False):
    logger = logging.getLogger('polipy')

    datefmt = '%Y-%m-%d %H:%M:%S'
    log_format = '[%(asctime)s] - %(levelname)-8s %(message)s'
    formatter = logging.Formatter(log_format, datefmt)

    stdout_handler = logging.StreamHandler(sys.stdout)
    stdout_handler.setFormatter(formatter)
    logger.addHandler(stdout_handler)
    logger.setLevel(logging.INFO) if verbose else logger.setLevel(logging.WARNING)
    return logger



# -*- coding: UTF-8 -*-

import argparse
from bs4 import BeautifulSoup
import datetime
import hashlib
import logging
import multiprocessing
import os
import selenium
import sys
import time
import pickle
import tldextract
import queue
import requests
import io

# from pyPdf import PdfFileReader
from tika import parser
from tqdm import tqdm
from selenium import webdriver
from selenium.webdriver.remote.command import Command

logger = logging.getLogger(__name__)
logger.addHandler(logging.StreamHandler(sys.stdout))

###############################################################################
# Define classes.
###############################################################################

class PrivacyPolicy:
    def __init__(self, url, is_active=0):
        self.url = url
        self.url_md5 = hashlib.md5(url.encode('utf-8')).hexdigest()
        self.is_active = is_active

        domain = tldextract.extract(url)
        self.domain = '.'.join(x for x in domain if x is not None).strip().strip('.') # Domains can't start or end with spaces or dots

    def __str__(self):
        return 'url=%s,domain=%s,url_md5=%s,is_active=%d' % (self.url, self.domain, self.url_md5, self.is_active)

    def __hash__(self):
        return hash(self.url_md5)

    def __eq__(self, other):
        return isinstance(other, self.__class__) and \
               self.url_md5 == other.url_md5

class ParallelArg:
    def __init__(self, policy, output_dir, utc_date=None, check_date=None, browser_language='en-US, en', verbose=False):
        self.policy = policy
        self.output_dir = output_dir
        self.check_date = check_date
        self.utc_date = utc_date
        self.browser_language = browser_language
        self.verbose = verbose

###############################################################################
# Setup and helper functions.
###############################################################################

def _parse_args():
    parser = argparse.ArgumentParser(description='Download privacy policies, optionally update the DB')
    parser.add_argument('input_path', help='Path to file where policy urls are located.')
    parser.add_argument('output_dir', help='Path to directory where policies will be saved. Creates directory structure <outputdir>/<date>/<regiontag>/<domain>/<urlhash>/')
    parser.add_argument('--processes', '-p', default=multiprocessing.cpu_count(), type=int, help='Number of processes to use')
    parser.add_argument('--check_previous', '-c', default=False, action='store_true', help='Boolean indicating whether to check against previous policies')
    parser.add_argument('--language', '-l', default='en-US, en', help='Language string to set in Firefox\'s intl.accept_languages option. Defaults to "en_US, en"')
    parser.add_argument('--verbose', '-v', action='store_true', help='Enable verbose logging')
    return parser.parse_args()

def _enable_verbose_logging(enable):
    logging.basicConfig()
    if(enable):
        logger.setLevel(logging.INFO)

def get_urls_from_file(input_file):
    # Get information about apps.
    with open(input_file, 'r') as f:
        policy_urls = f.read().split('\n')
    return policy_urls

def _write_file(content, output, write_mode='w'):
    with open(output, write_mode) as out_file:
        out_file.write(content)

###############################################################################
# Content extractors.
###############################################################################

def _extract_pdf(pdf_content):
    """
    From https://stackoverflow.com/questions/45470964/python-extracting-text-from-webpage-pdf
    """
    f = io.BytesIO(pdf_content)

    # First try using Tika parser, which gives better results.
    raw = parser.from_buffer(f)
    content = raw['content'] if 'content' in raw else ''

    # If that does not work, use PdfFileReader. Note: requires many dependencies.
    # reader = PdfFileReader(f)
    # if content is None or content == '':
    #     N = reader.getNumPages()
    #     content = [reader.getPage(i).extractText().strip() for i in range(N)]
    #     content = '\n'.join(content)

    return content.strip()

def _extract_google_docs(html):
    chunks = html.split('"s":"')[1:]
    chunks = [x.split('"},')[0].replace('\\n', '\n').replace('\\u000b', '\n') for x in chunks]
    text = ''.join(chunks).strip()
    return text

def _extract_text(html):
    """
    From https://stackoverflow.com/questions/328356/extracting-text-from-html-file-using-python
    """
    soup = BeautifulSoup(html, 'lxml')

    for script in soup(['script','style']):
        script.extract()

    text = soup.get_text(separator=' ')
    lines = (line.strip() for line in text.splitlines())
    chunks = (phrase.strip() for line in lines for phrase in line.split('  '))
    text = '\n'.join(chunk for chunk in chunks if chunk)

    return text

###############################################################################
# Zombie processes killers.
###############################################################################

def _kill_zombies_parallel_wrapper(q, zombie_timeout=300):
    while True:
        time.sleep(zombie_timeout)
        _kill_zombies(zombie_timeout=zombie_timeout)
        try:
            q.get_nowait()
            logger.info('Kill zombies queue no longer empty, terminating process')
            break
        except queue.Empty:
            continue

def _kill_zombies(zombie_timeout=5):
    if(os.name == 'posix'):
        logger.info('Killing hung geckodriver and firefox processes running for more than %d seconds' % zombie_timeout)
        os.system('ps -eo pid,ppid,etime,command | grep geckodriver | grep -v grep  | awk \'{if (int(substr($3,0,2)) > %d) {print $1,$2}}\' | xargs kill -9 || true > /dev/null 2>&1' % zombie_timeout)
        os.system('ps -eo pid,ppid,etime,command | grep "firefox -marionette" | grep -v grep  | awk \'{if (int(substr($3,0,2)) > %d) {print $1,$2}}\' | xargs kill -9 || true > /dev/null 2>&1' % zombie_timeout)
    else:
        logger.warning('Can\'t kill zombie processes on non-posix system "%s"' % os.name)

###############################################################################
# Main download policy function.
###############################################################################

def download_policy(policy, output_dir, check_date=None, utc_date=None, browser_language='en-US, en', browser_max_tries=10, browser_page_load_timeout=30):
    '''
    Output files:
    <domain>--<identifier>[--<region_tag>]--<utc_date>
    - .html: HTML content dump from URL
    - .txt: Text content extraction from HTML
    - .png: Full-screen content screenshot from URL
    - .url: Text file containing URL
    - .md5: MD5 hex digest of the raw file
    '''
    assert isinstance(policy, PrivacyPolicy), 'policy argument must be of type PrivacyPolicy'
    assert os.path.isdir(output_dir), 'Output directory %s does not exist' % output_dir

    if(utc_date is None):
        utc_date = datetime.datetime.utcnow().strftime('%Y%m%d')
        logger.info('utc_date was None, setting to %s' % utc_date)

    filename_parts = [policy.domain, policy.url_md5, utc_date]
    output_filename = '--'.join(x for x in filename_parts if x is not None)

    logger.info('Downloading privacy policy %s into %s' % (policy.url, output_dir))

    ff_driver = None
    try:
        ff_opts = webdriver.FirefoxOptions()
        ff_opts.add_argument('--headless')
        ff_opts.add_argument('--private')
        ff_prof = webdriver.FirefoxProfile()
        ff_prof.set_preference('intl.accept_languages', browser_language)
        while(ff_driver is None and browser_max_tries > 0):
            try:
                logger.info('Initializing Firefox web driver for %s' % policy.url)
                ff_driver = webdriver.Firefox(options=ff_opts, firefox_profile=ff_prof)
                logger.info('Firefox driver initialized for %s' % policy.url)
            except:
                browser_max_tries = browser_max_tries - 1
                logger.exception('Exception while creating Firefox driver for %s. Will try %d more times.' % (policy.url, browser_max_tries))
        assert ff_driver is not None, 'Firefox web driver failed to initialize for %s' % policy.url
        ff_driver.set_page_load_timeout(browser_page_load_timeout)
        logger.info('Getting %s' % policy.url)
        ff_driver.get(policy.url)
        logger.info('Got %s' % policy.url)

        # Extract text from documents.
        page_source = ff_driver.page_source
        if policy.url.endswith('.pdf'):
            response = requests.get(policy.url, verify=False)
            extracted_text = _extract_pdf(response.content)
        elif policy.domain == 'docs.google.com' and '"s":"' in page_source:
            extracted_text = _extract_google_docs(page_source)
        else:
            extracted_text = _extract_text(page_source)

        # Check if policy has been updated based on the text of the policy.
        check_dir, check_file_path = None, None
        while check_date is not None:
            filename_parts_past = [policy.domain, policy.url_md5, check_date]
            check_filename = '--'.join(x for x in filename_parts_past if x is not None)
            check_output_dir = os.path.abspath(os.path.join(os.path.abspath(output_dir), '..', '..', '..', check_date))
            check_dir = os.path.join(check_output_dir, policy.domain, policy.url_md5)

            if os.path.exists(os.path.join(check_dir, 'output.out')):
                with open(os.path.join(check_dir, 'output.out')) as date_file:
                    check_date = date_file.read()
            else:
                break

        # Check if the current policy matches the previous one, if possible.
        if check_dir is not None and os.path.exists(check_dir):
            check_output_file = '%s.txt' % check_filename
            check_file_path = os.path.join(check_dir, check_output_file)

        # If policy has not been updated, write an output.out file.
        if check_file_path is not None and os.path.exists(check_file_path):
            with open(check_file_path) as prev_txt:
                policy_updated = extracted_text != prev_txt.read()
            if policy_updated:
                logger.info('There is an updated privacy policy for %s - scraping', policy.url)
            else:
                logger.info('No difference found between previous policy for %s - exiting', policy.url)
                _write_file(check_date, os.path.join(output_dir, 'output.out'))
                return

        output_url = '%s.url' % output_filename
        logger.info('Saving %s as URL file %s' % (policy.url, output_url))
        _write_file(policy.url, os.path.join(output_dir, output_url))

        output_html = '%s.html' % output_filename
        logger.info('Saving %s as HTML dump %s' % (policy.url, output_html))
        _write_file(page_source, os.path.join(output_dir, output_html))

        output_txt = '%s.txt' % output_filename
        logger.info('Saving %s as text extraction %s' % (policy.url, output_txt))
        _write_file(extracted_text, os.path.join(output_dir, output_txt))

        output_md5 = '%s.md5' % output_filename
        logger.info('Saving %s\'s MD5 sum as %s' % (policy.url, output_md5))
        file_md5 = hashlib.md5(open(os.path.join(output_dir, output_html), 'rb').read()).hexdigest()
        _write_file(file_md5, os.path.join(output_dir, output_md5))

        output_png = '%s.png' % output_filename
        logger.info('Saving %s as PNG screenshot %s' % (policy.url, output_png))
        ff_html = ff_driver.find_element_by_tag_name('html')
        ff_html.screenshot(os.path.join(output_dir, output_png))

        if not policy.url.endswith('.pdf'):
            return

        output_pdf = '%s.pdf' % output_filename
        logger.info('Saving %s associated PDF file %s' % (policy.url, output_pdf))
        _write_file(response.content, os.path.join(output_dir, output_pdf), write_mode='wb')


    except Exception as e:
        logger.exception('Error while downloading policy.url: %s' % policy.url)

    finally:
        if(ff_driver is not None):
            try:
                logger.info('Closing Firefox web driver for %s' % policy.url)
                ff_driver.quit()
                logger.info('ff_driver.quit() success for %s' % policy.url)
                return
            except:
                logger.exception('Error while closing Firefox web driver for: %s' % policy.url)
        else:
            logger.warning('ff_driver is None for %s' % policy.url)

def _download_policy_parallel_wrapper(parallel_arg):
    logger = logging.getLogger(__name__)
    _enable_verbose_logging(parallel_arg.verbose)

    assert isinstance(parallel_arg, ParallelArg), 'Argument to _download_policy_parallel_wrapper must be of type ParallelArg'

    assert not os.path.isfile(parallel_arg.output_dir), 'Policy output directory %s already exists as a file'
    if(not os.path.isdir(parallel_arg.output_dir)):
        logger.info('Creating policy output directory %s' % parallel_arg.output_dir)
        os.makedirs(parallel_arg.output_dir)
    assert os.path.isdir(parallel_arg.output_dir), 'Policy output directory %s does not exist' % parallel_arg.output_dir

    # Only scrape if empty.
    file_num = len(os.listdir(parallel_arg.output_dir))
    if file_num > 0:
        logger.info('Policy output directory {} has {} files; skipping'.format(parallel_arg.output_dir, file_num))
        return

    download_policy(parallel_arg.policy, \
                    parallel_arg.output_dir, \
                    check_date=parallel_arg.check_date, \
                    utc_date=parallel_arg.utc_date, \
                    browser_language=parallel_arg.browser_language)

def _download_policies(policies, output_dir, language, verbose, check_previous, processes=multiprocessing.cpu_count()):
    unique_policies = set(policies)
    logger.info('Attempting to download %d policies' % len(unique_policies))
    logger.info('output_dir=%s, language=%s' % (output_dir, language))

    # Create the tagged output directory <output_dir>/<date>/<region_tag>/, if necessary
    utc_date = datetime.datetime.utcnow().strftime('%Y%m%d')
    tagged_output_dir = os.path.join(os.path.abspath(output_dir), utc_date) #, region_tag)
    assert not os.path.isfile(tagged_output_dir), 'Tagged output directory %s already exists as a file'
    if(not os.path.isdir(tagged_output_dir)):
        logger.info('Creating tagged output directory %s' % tagged_output_dir)
        os.makedirs(tagged_output_dir)
    assert os.path.isdir(tagged_output_dir), 'Tagged output directory %s does not exist' % tagged_output_dir

    # Set up the process to kill hung geckodrivers
    logger.info('Setting up geckodriver killer')
    queue = multiprocessing.Queue()
    killer_process = multiprocessing.Process(target=_kill_zombies_parallel_wrapper, args=(queue,))
    killer_process.start()

    # Calculate date to check by if update argument is true
    check_date = None
    if check_previous:
        directory_dates = []
        for d in os.listdir(os.path.abspath(output_dir)):
            if d != utc_date:
                try:
                    directory_dates.append(datetime.datetime.strptime(d, '%Y%m%d'))
                except ValueError:
                    pass
        check_date = max(directory_dates).strftime('%Y%m%d') if len(directory_dates) > 0 else None

    # Download the policies
    logger.info('Parallelizing downloads over %d processes' % processes)
    parallel_args = [ParallelArg(x, \
                                 os.path.join(tagged_output_dir, x.domain, x.url_md5), \
                                 utc_date=utc_date, \
                                 browser_language=language, \
                                 check_date=check_date, \
                                 verbose=verbose) \
                     for x in unique_policies]

    with multiprocessing.Pool(processes=processes) as pool:
        r = list(tqdm(pool.imap_unordered(_download_policy_parallel_wrapper, parallel_args), total=len(parallel_args)))

    # Kill the geckodriver killer process
    logger.info('Stopping the geckodriver killer process')
    queue.put(True)
    queue.close()
    queue.join_thread()
    killer_process.join()

###############################################################################
# Main function.
###############################################################################

if __name__ == '__main__':
    args = _parse_args()
    _enable_verbose_logging(enable=True)
    assert (args.output_dir is not None), 'Must specify output directory'
    assert (args.input_path is not None), 'Must specify file containing policy urls'

    processes = args.processes
    if processes is None:
        processes = 1
    assert (isinstance(processes, int) and processes > 0), 'Must specify positive number of processes'

    # Load privacy policies.
    policy_urls = get_urls_from_file(args.input_path)

    # Policy URLs should be list of PrivacyPolicy objects.
    policies = [PrivacyPolicy(url) for url in policy_urls]

    # Download policies
    _download_policies(policies, args.output_dir, args.language, args.verbose, args.check_previous, processes=processes)



import os
from setuptools import setup, find_packages

# The directory containing this file.
WORK_DIR = os.path.dirname(os.path.abspath(__file__))

# The text of the README file.
README = os.path.join(WORK_DIR, 'README.md')
with open(README, 'r') as f:
    long_description = f.read()

setup(
    name='polipy',
    version='0.1.2',
    description='Library for scraping, parsing, and analyzing privacy policies.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/blues-lab/polipy',
    author='Nikita Samarin',
    author_email='nsamarin@berkeley.edu',
    license='GNU General Public License v3.0 (GPLv3)',
    classifiers=[
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.9'
    ],
    packages=find_packages(),
    include_package_data=False,
    install_requires=[
        'beautifulsoup4',
        'certifi',
        'cffi',
        'chardet',
        'cryptography',
        'filelock',
        'idna',
        'pdfminer.six',
        'pycparser',
        'requests',
        'requests-file',
        'selenium',
        'six',
        'sortedcontainers',
        'soupsieve',
        'tqdm',
        'urllib3'
    ],
    entry_points={
        'console_scripts': [
            'polipy=polipy.__main__:main',
        ]
    },
)
